use clichess;
use clichess::{GameInfo, Player, RecvPositionError, UserRole, EXIT_MSG};
use serde_json::Value;
use shakmaty::{fen, Chess, Color, Setup};
use std::collections::HashMap;
use std::fs;
use std::io;
use std::os::unix::net::{UnixListener, UnixStream};
use std::sync::mpsc::{channel, Receiver, Sender, TryRecvError};
use std::sync::{Arc, Condvar, Mutex};

use std::thread;

struct Server {
    id: usize,
    chess_position: Arc<Mutex<Chess>>,
    players: Arc<Mutex<HashMap<usize, (Player, Sender<GameInfo>, Arc<(Mutex<bool>, Condvar)>)>>>,
    others_serv_msg_recv: Receiver<GameInfo>,
    client_message_recv: Receiver<String>,
    cvar: Arc<(Mutex<bool>, Condvar)>,
    opponent_name: Option<String>,
}

fn main() {
    let chess = Arc::new(Mutex::new(Chess::default()));
    let players = Arc::new(Mutex::new(HashMap::new()));
    let mut counter = 0;
    fs::remove_file("/tmp/clichess.socket");

    let listener = UnixListener::bind("/tmp/clichess.socket").unwrap();

    // accept connections and process them, spawning a new thread for each one
    for stream in listener.incoming() {
        match stream {
            Ok(stream) => {
                let condvar_pair = Arc::new((Mutex::new(false), Condvar::new()));
                let client_message_recv =
                    setup_client_message_recv(&stream, condvar_pair.clone()).unwrap();
                let (others_serv_msg_sender, others_serv_msg_recv) = channel();
                let mut server = Server {
                    id: counter,
                    chess_position: chess.clone(),
                    players: players.clone(),
                    others_serv_msg_recv,
                    client_message_recv,
                    cvar: condvar_pair,
                    opponent_name: Option::None,
                };
                /* connection succeeded */
                thread::spawn(move || handle_player(stream, &mut server, others_serv_msg_sender));
                counter += 1;
            }
            Err(err) => {
                println!("Could not create a connection: {}", err);
                break;
            }
        }
    }
}

fn handle_player(
    mut stream: UnixStream,
    server: &mut Server,
    others_serv_msg_sender: Sender<GameInfo>,
) {
    match initialize_client(&mut stream, &server, others_serv_msg_sender) {
        Ok((player, player_turn)) => main_loop(&mut stream, server, player, player_turn),
        Err(e) => println!("User id {} could not be initialized: {}", server.id, e),
    };
    player_disconnected(&server);
}

fn initialize_client(
    stream: &mut UnixStream,
    server: &Server,
    others_serv_msg_sender: Sender<GameInfo>,
) -> Result<(Player, Color), RecvPositionError> {
    //create player
    let player = create_player(server, stream, others_serv_msg_sender)?;
    let player_turn: Color;
    //send current position to the player
    println!(
        "server {}, send current position to the player...",
        server.id
    );
    {
        let chess = server.chess_position.lock().unwrap();
        player_turn = chess.turn();
        clichess::write_to_stream(stream, fen::fen(&*chess)).unwrap();
    }
    println!("server {}, current position to the player sent", server.id);
    Ok((player, player_turn))
}

fn main_loop(stream: &mut UnixStream, server: &mut Server, player: Player, mut player_turn: Color) {
    loop {
        if clichess::is_player_turn(&player, player_turn) {
            //let go of the lock while waiting for user input.
            println!("server {}, waiting for player move..", server.id);
            let input;
            match server.client_message_recv.recv() {
                Ok(i) => input = i,
                Err(e) => {
                    println!("Error while getting user input: {}", e);
                    break;
                }
            };
            if input == EXIT_MSG {
                break;
            }
            {
                let mut chess = server.chess_position.lock().unwrap();
                let players = server.players.lock().unwrap();
                println!("User tried to play {}", input);
                match clichess::try_to_play_move(&chess, input) {
                    Ok(played_chess) => *chess = played_chess,
                    Err(e) => println!("Error: {}", e),
                };
                let game_info_to_send = GameInfo {
                    game_fen: fen::fen(&*chess),
                    opponent_name: players
                        .get(&server.id)
                        .expect("current server is in the server list")
                        .0
                        .username
                        .clone(),
                };
                for (id, (_, others_serv_msg_sender, cvar_pair)) in players.iter() {
                    if server.id != *id {
                        others_serv_msg_sender
                            .send(game_info_to_send.clone())
                            .unwrap();
                        let (lock, cvar) = &**cvar_pair;
                        let mut message_sent = lock.lock().unwrap();
                        *message_sent = true;
                        cvar.notify_one();
                    }
                }
                if clichess::write_to_stream(
                    stream,
                    serde_json::to_string(&GameInfo {
                        game_fen: fen::fen(&*chess),
                        opponent_name: server.opponent_name.clone().unwrap_or_default(),
                    })
                    .unwrap(),
                )
                .is_err()
                {
                    break;
                }
                player_turn = chess.turn();
            }
        } else {
            let game_info_result = wait_for_opponent_move(server);
            if game_info_result.is_err() {
                break;
            }
            let game_info = game_info_result.unwrap();
            clichess::write_to_stream(stream, serde_json::to_string(&game_info).unwrap()).unwrap();
            server.opponent_name = Some(game_info.opponent_name);
            let chess = server.chess_position.lock().unwrap();
            player_turn = chess.turn();
        }
    }
}

fn wait_for_opponent_move(server: &Server) -> Result<GameInfo, RecvPositionError> {
    println!("server id: {}, waiting for move to send...", server.id);
    //wait: either we receive next position from other server threads, or we receive
    //"exit" from the client.
    let mut returned_result = Err(RecvPositionError::UserCanceledError);
    let (lock, cvar) = &*server.cvar;
    let mut message_sent = lock.lock().unwrap();
    *message_sent = false;
    while !*message_sent {
        message_sent = cvar.wait(message_sent).unwrap();
    }
    *message_sent = false;
    match server.others_serv_msg_recv.try_recv() {
        Ok(msg) => {
            returned_result = Ok(msg);
        }
        Err(TryRecvError::Disconnected) => {
            println!("Error: other server thread disconnected while sending move.")
        }
        Err(TryRecvError::Empty) => { /*nothing to do*/ }
    }
    match server.client_message_recv.try_recv() {
        Ok(msg) => {
            if msg == EXIT_MSG {
                returned_result = Err(RecvPositionError::UserCanceledError);
            } else {
                println!("Client sent message while it's not its turn, this is an error.");
                println!("Here is the message: {}", msg);
                returned_result = Err(RecvPositionError::UserCanceledError);
            }
        }
        Err(TryRecvError::Disconnected) => println!("Error: client disconnected."),
        Err(TryRecvError::Empty) => {}
    }
    returned_result
}

fn create_player(
    server: &Server,
    stream: &mut UnixStream,
    others_serv_msg_sender: Sender<GameInfo>,
) -> Result<Player, RecvPositionError> {
    println!("Creating player {}...", server.id);
    //get player name and pubkey
    let username_pubkey_json = server
        .client_message_recv
        .recv()
        .expect("Player closed connection.");
    let username_pubkey_value: Value = serde_json::from_str(&username_pubkey_json).unwrap();
    let username = username_pubkey_value["username"].to_string();
    let public_key = username_pubkey_value["pubkey"].to_string();
    println!("got username: {}", username);
    println!("got pubkey: {}", public_key);
    let role = receive_user_role(server, stream)?;
    let mut players = server.players.lock().unwrap();
    let player = Player {
        role,
        username,
        public_key,
    };
    players.insert(
        server.id,
        (player.clone(), others_serv_msg_sender, server.cvar.clone()),
    );
    println!("Created player {}", server.id);
    Ok(player)
}

fn receive_user_role(
    server: &Server,
    stream: &mut UnixStream,
) -> Result<UserRole, RecvPositionError> {
    let mut chosen_role: UserRole;
    loop {
        //send available roles
        let available_roles = compute_available_roles_to_str(server);
        println!("Computed available_roles as str: {}", available_roles);
        clichess::write_to_stream(stream, available_roles.clone()).unwrap();
        //receive chosen role
        let chosen_role_str = server
            .client_message_recv
            .recv()
            .expect("Player closed connection.");
        if chosen_role_str.trim() == EXIT_MSG {
            return Err(RecvPositionError::UserCanceledError);
        }
        println!("Client id {} has chosen {}", server.id, chosen_role_str);
        let chosen_role_parse = clichess::parse_to_role(&chosen_role_str);

        if chosen_role_parse.is_err() {
            return Err(RecvPositionError::ParsePositionError);
        }
        chosen_role = chosen_role_parse.unwrap();
        //check if role is still available after the choice
        {
            let mut available_roles_after_choice =
                vec![UserRole::White, UserRole::Black, UserRole::Spectator];
            let players = server.players.lock().unwrap();
            for (_, (player, _, _)) in players.iter() {
                match available_roles_after_choice
                    .iter()
                    .position(|r| *r == player.role)
                {
                    Some(index) => available_roles_after_choice.remove(index),
                    None => continue,
                };
            }
            if available_roles_after_choice.contains(&chosen_role) {
                println!("OK");
                clichess::write_to_stream(stream, String::from("OK")).unwrap();
                server
                    .client_message_recv
                    .recv()
                    .expect("Player closed connection.");
                break;
            } else {
                println!("KO");
                clichess::write_to_stream(stream, String::from("KO")).unwrap();
                server
                    .client_message_recv
                    .recv()
                    .expect("Player closed connection.");
                continue;
            }
        }
    }
    Ok(chosen_role)
}

fn compute_available_roles_to_str(server: &Server) -> String {
    let mut available_roles = vec![UserRole::White, UserRole::Black, UserRole::Spectator];
    let players = server.players.lock().unwrap();
    for (_, (player, _, _)) in players.iter() {
        match available_roles.iter().position(|r| *r == player.role) {
            Some(index) => available_roles.remove(index),
            None => continue,
        };
    }
    let available_roles_str: Vec<String> = available_roles
        .iter()
        .map(|r| clichess::role_to_str(r))
        .collect();
    available_roles_str.join(",")
}

fn player_disconnected(server: &Server) {
    let mut players = server.players.lock().unwrap();
    let player_and_receiving_buf = players.get(&server.id);
    if player_and_receiving_buf.is_some() {
        let (player, _, _) = player_and_receiving_buf.expect("is some");
        println!("Player {} disconnected.", player.username);
    }
    players.remove(&server.id);
}

fn setup_client_message_recv(
    stream: &UnixStream,
    cvar_pair: Arc<(Mutex<bool>, Condvar)>,
) -> io::Result<Receiver<String>> {
    let (sender, receiver) = channel();
    let thread_stream = stream.try_clone()?;

    thread::spawn(move || {
        loop {
            //wait for client message
            let buffer =
                clichess::read_line_from_stream(&thread_stream).expect("Error message from server");
            //notify the server if waiting
            let (lock, cvar) = &*cvar_pair;
            let mut message_sent = lock.lock().unwrap();
            *message_sent = true;
            cvar.notify_one();
            let send_result = sender.send(buffer);
            // stop the thread during a disconnection
            if send_result.is_err() {
                break;
            }
        }
    });
    Ok(receiver)
}
