use colored::Colorize;
use serde::{Deserialize, Serialize};
use shakmaty::san::ParseSanError;
use shakmaty::san::San;
use shakmaty::san::SanError;
use shakmaty::uci::Uci;
use shakmaty::{Chess, Color, IllegalMoveError, Position, Setup, Square};
use std::fmt;
use std::io;
use std::io::prelude::*;
use std::io::{BufReader, Write};
use std::os::unix::net::UnixStream;

//message to send to the server to signal we disconnected.
pub const EXIT_MSG: &str = "exit";

#[derive(Clone)]
pub struct Player {
    pub role: UserRole,
    pub username: String,
    pub public_key: String,
}

#[derive(Serialize, Deserialize, Clone)]
pub struct GameInfo {
    pub game_fen: String,
    pub opponent_name: String,
}

#[derive(PartialEq, Copy, Clone)]
pub enum UserRole {
    White,
    Black,
    Spectator,
}

impl fmt::Display for UserRole {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                UserRole::White => "white",
                UserRole::Black => "black",
                UserRole::Spectator => "spectator",
            }
        )
    }
}

#[derive(Debug)]
pub enum RecvPositionError {
    UserCanceledError,
    CommunicationError,
    ParsePositionError,
}

impl fmt::Display for RecvPositionError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                RecvPositionError::CommunicationError =>
                    "Error during the communication between client and server.",
                RecvPositionError::ParsePositionError => "Error when parsing current position.",
                RecvPositionError::UserCanceledError => "Cancelled by user.",
            }
        )
    }
}

struct SquareToPrint {
    color: String,
    background_color: String,
    square_representation: String,
}

#[derive(Debug)]
pub enum MoveInputError {
    SanSanError(SanError),
    SanParseSanError(ParseSanError),
    ShakmatyIllegalMoveError(IllegalMoveError),
}

impl From<SanError> for MoveInputError {
    fn from(error: SanError) -> Self {
        MoveInputError::SanSanError(error)
    }
}

impl From<ParseSanError> for MoveInputError {
    fn from(error: ParseSanError) -> Self {
        MoveInputError::SanParseSanError(error)
    }
}

impl From<IllegalMoveError> for MoveInputError {
    fn from(error: IllegalMoveError) -> Self {
        MoveInputError::ShakmatyIllegalMoveError(error)
    }
}

impl fmt::Display for MoveInputError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            MoveInputError::SanSanError(san_error) => san_error.fmt(f),
            MoveInputError::SanParseSanError(parse_san_error) => parse_san_error.fmt(f),
            MoveInputError::ShakmatyIllegalMoveError(illegal_move_error) => {
                illegal_move_error.fmt(f)
            }
        }
    }
}

pub fn board_representation(chess: &Chess, side: Color) -> String {
    let mut s = String::from("");
    let mut rank_numbers;
    if side == Color::White {
        rank_numbers = 8;
    } else {
        rank_numbers = 1;
    }
    for v in board_string_representation(chess, side) {
        for square_to_print in v {
            s.push_str(&format!(
                "{}",
                square_to_print
                    .square_representation
                    .color(square_to_print.color)
                    .on_color(square_to_print.background_color)
            ));
        }
        s.push_str(&format!("{}\n\r", rank_numbers));
        if side == Color::White {
            rank_numbers -= 1;
        } else {
            rank_numbers += 1;
        }
    }
    let mut files = vec!["a", "b", "c", "d", "e", "f", "g", "h"];
    if side == Color::Black {
        files.reverse();
    }
    for file in files {
        s.push_str(&format!("{} ", file));
    }
    s.push_str("\n\r");
    s
}

fn board_string_representation(chess: &Chess, side: Color) -> Vec<Vec<SquareToPrint>> {
    let mut full_board_to_print = Vec::new();
    for _ in 0..8 {
        full_board_to_print.push(Vec::new())
    }
    for i in 0..8 {
        for j in 0..8 {
            let square = Square::new(i + 8 * j);
            let square_to_print = SquareToPrint {
                color: "black".to_string(),
                background_color: if square.is_light() {
                    "white".to_string()
                } else {
                    "green".to_string()
                },
                square_representation: get_square_representation(&square, chess),
            };
            full_board_to_print[j as usize].push(square_to_print);
        }
    }
    if side == Color::White {
        full_board_to_print.reverse();
    } else {
        for j in 0..8 {
            full_board_to_print[j as usize].reverse();
        }
    }
    full_board_to_print
}

pub fn try_to_play_move(chess: &Chess, movestr: String) -> Result<Chess, MoveInputError> {
    //if uci parse error, try to parse with san.
    let mv = match movestr.parse::<Uci>() {
        Ok(uci) => uci.to_move(chess)?,
        Err(_) => movestr.parse::<San>()?.to_move(chess)?,
    };
    Ok((*chess).clone().play(&mv)?)
}

fn get_square_representation(square: &Square, chess: &Chess) -> String {
    let board = (*chess).board();
    match board.piece_at(*square) {
        Some(piece) => format!("{} ", piece_char_to_utf8(piece.char())),
        None => "  ".to_string(),
    }
}

fn piece_char_to_utf8(piece: char) -> char {
    match piece {
        'P' => '♙',
        'p' => '♟',
        'N' => '♘',
        'n' => '♞',
        'B' => '♗',
        'b' => '♝',
        'R' => '♖',
        'r' => '♜',
        'Q' => '♕',
        'q' => '♛',
        'K' => '♔',
        'k' => '♚',
        c => c,
    }
}

pub fn read_line_from_stream(stream: &UnixStream) -> Result<String, io::Error> {
    let mut result = String::new();
    let mut reader = BufReader::new(stream);

    reader.read_line(&mut result)?;
    Ok(String::from(result.trim_end()))
}

pub fn write_to_stream(stream: &mut UnixStream, msg: String) -> io::Result<()> {
    stream.write_all(&(msg + "\n").as_bytes())?;
    Ok(())
}

pub fn get_default_side(role: UserRole) -> Color {
    if role == UserRole::Black {
        Color::Black
    } else {
        Color::White
    }
}

pub fn parse_to_role(s: &str) -> Result<UserRole, String> {
    match &*s.to_ascii_lowercase() {
        "w" => Ok(UserRole::White),
        "b" => Ok(UserRole::Black),
        "s" => Ok(UserRole::Spectator),
        _ => Err(String::from("Please enter a valid answer.")),
    }
}

pub fn role_to_str(r: &UserRole) -> String {
    String::from(match *r {
        UserRole::White => "w",
        UserRole::Black => "b",
        UserRole::Spectator => "s",
    })
}

pub fn is_player_turn(player: &Player, playing_side: Color) -> bool {
    (playing_side == Color::White && player.role == UserRole::White)
        || (playing_side == Color::Black && player.role == UserRole::Black)
}
